package Warsync;

##############################################################################
## Warsync.pm - Warsync                                                     ##
##                                                                          ##
## Copyright (C) 2001-2004 Paul J. Baker                                    ##
##                                                                          ##
## This program is free software; you can redistribute it and/or modify     ##
## it under the terms of the GNU General Public License version 2 as        ##
## published by the Free Software Foundation.                               ##
##                                                                          ##
## This program is distributed in the hope that it will be useful,          ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            ##
## GNU General Public License for more details.                             ##
##                                                                          ##
## You should have received a copy of the GNU General Public License        ##
## along with this program; if not, write to the Free Software              ##
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                ##
##                                                                          ##
## You may contact the author via electronic mail at pfars@paulbaker.net    ##
##############################################################################

## $Id: Warsync.pm,v 1.23 2004/03/27 22:45:12 pbaker-guest Exp $

=head1 NAME

Warsync - Wrapper Around Rsync

=head1 SYNOPSIS

  use Warsync;

=head1 DESCRIPTION

Warsync exports various functions used by B<Warsync> scripts.

=head1 AUTHOR

Paul Baker, pbaker@where2getit.com

=head1 SEE ALSO

warsync(8).

=head1 VERSION

0.9.7

=cut

#############################################################################

require 5.6.0;
use strict;
use warnings;

use Carp;
use Data::Dumper;
use Getopt::Long ();
use Pod::Usage;
use Sys::Hostname;

use Warsync::Server;
use Warsync::Client;
use Warsync::Config;

require Exporter;

our @ISA = qw( Exporter );

our @EXPORT = qw( deep_copy fault is_ipaddr is_fqdn );

our $VERSION = '0.9.7';

## set option defaults.
Getopt::Long::Configure("bundling");

sub options
{
    my ($class, @args) = @_;

    my %option = ();

    Getopt::Long::GetOptions(
        \%option,
        'help|H|?',
        'man',
        'version|V',
        'v|verbose+',
        'q|quiet',
        'n|dry-run',
        @args
    ) or pod2usage(2);

    pod2usage(1) if $option{help};
    pod2usage(-exitstatus => 0, -verbose => 2) if $option{man};

    # display version number if --version set.
    if ($option{version}) {
        print "Warsync - Wrapper Around Rsync\n";
        print "Version $VERSION\n";
        exit(0);
    }

    return \%option;
}

sub fault ($;$) {
    print STDERR 'Error: ', $_[0], "\n";
    exit ($_[1] || 1);
}

sub prompt
{
    my ($class, %args) = @_;

    my $response = '';
    my @choices;

    if (exists $args{default}) {
        $response = $args{default};
    }
    
    if (exists $args{choices}) {
        $response ||= $args{choices}[0];
        @choices  = @{$args{choices}};
    }

    for (;;) {
        ## print the question.
        print $args{prompt};

        ## print list of choices.
        if (@choices) {
            print '(';
            print join( '/', map { (lc($response) eq lc($_))
                                   ? uc($_) : $_ } @choices );
            print ') ';
        }
        ## or print the default.
        elsif ($response) {
            print '[', $response, '] ';
        }
        ## mark end of prompt.
        print '>> ';
        ## get the response.
        chomp(my $in = <STDIN>);
        ## use if not blank.
        $response = $in if length($in);

        ## check that it matches a choice
        if (@choices) {
            foreach my $choice (@choices) {
                ## return if it matches.
                return lc($response) if lc($response) eq lc($choice);
            }
        }
        ## no specific choices, return any response.
        elsif ($response) {
            return $response;
        }
    }
}

sub is_ipaddr ($) {
    my $in = shift;

    my ($ip) = $in =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/;

    return $ip;
}


{
    ## hostname/domain regex
    my $rfc1035_label_re  = qr/[0-9A-Za-z](?:[-0-9A-Za-z]*[0-9A-Za-z])?/;

    sub is_fqdn ($) {
        my $in = shift || '';

        my ($hostname) = $in =~ /($rfc1035_label_re(?:\.$rfc1035_label_re)*)/;

        return $hostname;
    }
}

sub deep_copy ($);
sub deep_copy ($)
{
    ## http://www.stonehenge.com/merlyn/UnixReview/col30.html
    my $this = shift;
    if (not ref $this) {
        $this;
    }
    elsif (ref $this eq "ARRAY") {
        [map deep_copy($_), @$this];
    }
    elsif (ref $this eq "HASH") {
        +{map { $_ => deep_copy($this->{$_}) } keys %$this};
    }
    else {
        die "what type is $_?"
    }
}

sub parse_package_list ($)
{
    my $dpkg = shift;
    my %packages = ();

    while (my $line = <$dpkg>) {
        ## skip header lines.
        next if $line =~ /^[|+]|Desired/;
        chomp($line);

        if (my @pkg = $line =~
                /^
                (...) \s+              ## 0 - desired state
                ([-a-z0-9+.]+) \s+     ## 1 - package
                ([-.0-9+a-zA-Z:]+) \s+ ## 2 - version
                (\S.*)                 ## 3 - description
                /x)
        {
            $packages{$pkg[1]} = \@pkg;
        }
    }

    return \%packages;
}

sub debian_packages (;$)
{
    my $v = shift; ## verbose

    ## get current package information.
    my $packages;
    
    {
        my $prev_col = $ENV{COLUMNS};
        ## now set it to 200
        $ENV{COLUMNS} = 200;
    
        ## read dpkg -l
        open(my $dpkg, 'dpkg -l |')
            or die "could not exec dpkg -l: $!";
    
        $packages = parse_package_list($dpkg);
    
        close($dpkg);
    
        if ($prev_col) {
            $ENV{COLUMNS} = $prev_col;
        }
        else {
            delete $ENV{COLUMNS};
        }
    }

    my $total   = keys(%$packages);
    my $done    = 0;
    my $percent = 0;
    
    ## iterate through packages and get real version number.
    {
        open(my $fh, "</var/lib/dpkg/status")
            or fault "Could not read /var/lib/dpkg/status: $!";
        
        my $package = '';
        while (<$fh>) {
            chomp;
            next unless length;
            
            ## current package block.
            if (my ($pkg) = $_ =~ /^Package:\s+(\S+)/) {
                $package = $pkg;
                next;
            }
            
            next unless exists $packages->{$package};
            
            if (my ($ver) = $_ =~ /^Version:\s+(\S+)/) {
                $packages->{$package}[2] = $ver;
            
                if ($v + -t > 1) {
                    my $new = int( (++$done / $total) * 100);
                    if ($new > $percent) {
                        my $str = "$new%";
                        print STDERR "$str", "\b" x length($str);
                        $percent = $new;
                    } 
                }
            }
        }
        
        close($fh);
    }
    
    if ($v and -t) {
        print STDERR ' ' x 5, "\b" x 5;
    }

    return $packages;
}

sub debian_snapshot ($)
{
    my $option = shift;
    
    ## open file to write snapshot to.
    my $fn = $option->{'debian-snapshot'};
    open(my $fh, ">$fn") or fault "can not write to $fn: $!";
    
    print STDERR "Creating snapshot of Debian package installation...";
    
    my $packages = debian_packages(1);
    
    ## write package information to file.
    foreach my $pkg (sort keys %$packages) {
        my $p = $packages->{$pkg};
        print $fh $p->[0], ' ', $p->[1], ' ',
                  $p->[2], ' ', $p->[3], "\n";
    }
    
    close($fh);

    print "done.\n";

}
    
sub debian_package_diff ($$$;$)
{
    my $spkgs  = shift;
    my $cpkgs  = shift;
    my $ipkgs = shift;
    
    my $v     = shift || 0;
    
    my @install = ();
    my @remove  = ();
    my @purge   = ();

    my $total   = keys(%$spkgs) + keys(%$cpkgs);
    my $done    = 0;
    my $percent = 0;

    ## compare server packages to the client and create
    ## list of apt-get commands to run on the client.
    foreach my $pkg (keys %$spkgs) {

        ## display progress update.
        if (-t) {
            my $new = int( ($done++ / $total) * 100);
            if ($new > $percent) {
                my $str = "$new%";
                print "$str", "\b" x length($str);
                $percent = $new;
            } 
        }

        ## check if we should ignore this package.
        my $ignore = 0;
        foreach my $regex (@$ipkgs) {
            if ($pkg =~ /$regex/) {
                $ignore = 1;
                last;
            }
        }
        next if $ignore;

        my $spkg = $spkgs->{$pkg};

        ## if installed, make sure client installs it.
        if ($spkg->[0] eq 'ii ') {
            ## client has package installed
            if (exists $cpkgs->{$pkg} and $cpkgs->{$pkg}[0] eq 'ii ') {
                ## check version.
                if ($spkg->[2] ne $cpkgs->{$pkg}[2]) {
                    my $ver = $spkg->[2];
                    push @install, "$pkg=$ver";
                }
            }
            ## client does not have package installed.
            else {
                my $ver = $spkg->[2];
                push @install, "$pkg=$ver";
            }
        }

        ## if removed, make sure client removes it.
        elsif ($spkg->[0] =~ /^[rp]c $/) {
            ## client has package installed
            if (exists $cpkgs->{$pkg} and $cpkgs->{$pkg}[0] !~ /^[rp]c $/) {
                push @remove, $pkg;
            }
        }

        ## unexpected state, possibly flagged for removal,
        ## but not yet uninstalled.
        else {
            print STDERR "warn: server package $pkg $spkg->[2] has unexpected state: $spkg->[0].\n";
        }

        ## mark package as processed.
        $spkg->[5] = $cpkgs->{$pkg}[5] = 1;

    }

    ## now compare remaining client packages to the server.
    foreach my $pkg (keys %$cpkgs) {

        ## display progress update.
        if (-t) {
            my $new = int( ($done++ / $total) * 100);
            if ($new > $percent) {
                my $str = "$new%";
                print "$str", "\b" x length($str);
                $percent = $new;
            }
        }

        ## skip if we already processed it.
        next if $cpkgs->{$pkg}[5];

        ## check if we should ignore this package.
        my $ignore = 0;
        foreach my $regex (@$ipkgs) {
            if ($pkg =~ /$regex/) {
                $ignore = 1;
                last;
            }
        }
        next if $ignore;

        #print STDERR "client has package $pkg installed. server does not.\n";
        ## it shouldn't be here if it's not on the server.
        push @purge, $pkg;
    }
    
    return \@install, \@remove, \@purge;
}

sub debian_from_snapshot ($)
{
    my $option = shift;
    
    ## open file to read snapshot from.
    my $fn = $option->{'debian-from-snapshot'};
    open(my $fh, "<$fn") or fault "can not read from $fn: $!";
    my $spkgs = parse_package_list($fh);
    close($fh);
    
    print STDERR "Determining current package list...";
    my $cpkgs = debian_packages(1);
    
    print STDERR "done.\nCalculating differences...";
    
    my ($install, $remove, $purge) = debian_package_diff($spkgs, $cpkgs, [], 1);
    
    unless (@$install || @$remove || @$purge) {
        print STDERR "No differences.\n";
        return;
    }
    
    print STDERR "out-of-sync.\n";
    
    my $dryrun = $option->{n} ? ' (dry-run) ' : '';
    
    print STDERR "Installing/removing packages with apt-get$dryrun...\n";
        
    my @apt = 'apt-get';
    push @apt, '--dry-run' if $dryrun;
    push @apt, 'install', @$install;
    push @apt, map { "$_-" } @$remove;
    push @apt, map { "$_-" } @$purge;
        
    print STDERR "dry-run: @apt\n" if $dryrun;
        
    system(@apt) and fault("something went wrong running apt-get install: $?");
    
    if (@$purge) {
        print STDERR "Purging package configurations$dryrun...\n";
        
        my @dpkg = 'dpkg';
        push @dpkg, '--no-act' if $dryrun;
        push @dpkg, '--purge', @$purge;
        
        print STDERR "dry-run: @dpkg\n" if $dryrun;
        
        system(@dpkg) and fault("something went wrong running dpkg --purge: $?");
    }
    
    print STDERR "Debian package installation syncronized.\n";
}

1;
