package Warsync::Client;

use Warsync;

sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;

    ## only return object if this is a client system.
    ## check that we are a client by looking for the client
    ## authentication key and config file.
    unless (-r '/etc/warsync/client-key' and -r '/etc/warsync/client.conf') {
        return undef;
    }

    ## okay now parse the client config file.
    open(my $conf_fh, '/etc/warsync/client.conf')
        or Warsync::fault("could not open client config file: $!");

    $self = bless {} => $class;

    while (<$conf_fh>) {
        chomp;                  # no newline
        s/#.*//;                # no comments
        s/^\s+//;               # no leading white
        s/\s+$//;               # no trailing white
        next unless length;     # anything left?
        my ($var, $value) = split(/\s*=\s*/, $_, 2);
        $self->{ lc($var) } = $value;
    }

    close($conf_fh);

    ## must have a server.
    unless($self->{server} = Warsync::is_fqdn($self->{server})) {
        Warsync::fault("no server configured.");
    }

    return $self;
}

sub compress
{
    my $self = shift->new;
    
    if (@_) {
        $self->{compress} = shift;
        return $self;
    }
    
    return $self->{compress};
}

sub option
{
    my $self = shift->new;
    my $key  = shift;
    
    $self->options->{$key};
}

sub options
{
    my $self = shift->new;
    
    if (@_) {
        $self->{options} = shift;
        return $self;
    }
    return $self->{options};
}

sub package_install
{
    my $self     = shift->new;
    my $packages = shift;      ## space separated list of packages

    my @packages = split / /, $packages;

    my @apt = 'apt-get';
    push @apt, '--dry-run' if $self->option('n');

    print STDERR "dry-run: apt-get install @packages\n" if $self->option('n');

    push @apt, '-q', '-y' unless -t;
    
    system(@apt, 'install', @packages)
        and Warsync::fault("something went wrong running apt-get install: $?");
}

sub package_purge
{
    my $self     = shift->new;
    my $packages = shift;      ## space separated list of packages

    my @packages = split / /, $packages;

    my $option = $self->options;

    my @dpkg = 'dpkg';
    push @dpkg, '--no-act' if $self->option('n');

    print STDERR "dry-run: purge @packages\n" if $self->option('n');
    
    system(@dpkg, '--purge', @packages)
            and Warsync::fault("something went wrong running dpkg --purge: $?");
}

sub name
{
    my $self = shift->new;

    return Sys::Hostname::hostname;
}

sub packages
{
    my $self = shift->new;
    my $v    = shift;

    return Warsync::deep_copy(
        $self->{_packages} ||= Warsync::debian_packages($v)
    );
}

sub package_list
{
    my $self = shift->new;
    my $v    = shift;

    my $packages = $self->packages($v);
    
    my $list = '';
    
    foreach my $pkg (sort keys %$packages) {
        my $p = $packages->{$pkg};
        $list .= $p->[0] . ' ' . $p->[1] . ' '
               . $p->[2] . ' ' . $p->[3] . "\n";
    }
    
    return $list;
}

sub public_key
{
    my $self = shift->new;

    return `cat /etc/warsync/client-key.pub`;
}

sub run
{
    my $self = shift->new;

    my $option = Warsync->options(
        'debian-snapshot=s',
        'debian-from-snapshot=s',
    );

    $self->options($option);
    
    ## if --debian-snapshot argument given, do that.
    if ($option->{'debian-snapshot'}) {
        Warsync::debian_snapshot($option);
        exit;
    }
    
    ## if --debian-from-snapshot argument given, do that.
    if ($option->{'debian-from-snapshot'}) {
        Warsync::debian_from_snapshot($option);
        exit;
    }

}

sub server
{
    my $self = shift->new;

    return $self->{server};
}

sub server_port
{
    my $self = shift->new;
    
    return $self->{port} || 22;
}

sub server_command
{
    my $self = shift->new;
    my @cmd  = @_;

    my @scmd = ( 'ssh', '-p', $self->server_port,
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/client-key'    );
    
    push @scmd, '-C' if $self->compress;
    
    push @scmd, $self->server, @cmd;

    open(my $ssh_h, '-|') || exec( @scmd )
        or die "error '$!' trying to exec: @scmd";

    return $ssh_h;
}

sub server_name
{
    my $self = shift->new;

    my $server_h = $self->server_command('warsync-echo-name');

    my @scom = <$server_h>;

    close($server_h);

    if ($scom[0]) {
        my ($name) = $scom[0] =~ /^name: (\S+)/;
        return $name if $name;
    }

    Warsync::fault("could not retrieve server name.");
}

sub sync_package_sources
{
    my $self = shift->new;
    
    ## sync apt config.
    {
        my $sch = $self->server_command('warsync-echo-apt-conf');
        my @data = <$sch>;

        close($sch) or Warsync::fault("could not retrieve apt.conf: $?");

        if ($self->option('n')) {
            print STDERR "dry-run: syncronize /etc/apt/apt.conf\n";
        }
        else {
            open(my $fh, '> /etc/apt/apt.conf')
                or Warsync::fault("could not write apt.conf: $!");
            print $fh @data;
            close($fh);
        }
    }

    ## sync apt preferences.
    {
        my $sch = $self->server_command('warsync-echo-apt-preferences');
        my @data = <$sch>;

        close($sch) or Warsync::fault("could not retrieve preferences: $?");

        if ($self->option('n')) {
            print STDERR "dry-run: syncronize /etc/apt/preferences\n";
        }
        else {
            open(my $fh, '> /etc/apt/preferences')
                or Warsync::fault("could not write preferences: $!");
            print $fh @data;
            close($fh);
        }
    }

    ## sync apt sources.list.
    {
        my $sch = $self->server_command('warsync-echo-apt-sources-list');
        my @data = <$sch>;

        close($sch) or Warsync::fault("could not retrieve sources.list: $?");

        if ($option->{n}) {
            print STDERR "dry-run: syncronize /etc/apt/sources.list\n";
        }
        else {
            open(my $fh, '> /etc/apt/sources.list')
                or Warsync::fault("could not write to sources.list: $!");
            print $fh @data;
            close($fh);
        }
    }

    ## apt-get update
    {
        if ($self->option('n')) {
            print STDERR "dry-run: execute apt-get update\n";
        }
        elsif (-t) {
            system('apt-get', 'update')
                and Warsync::fault("something went wrong running apt-get update: $?");
        }
        else {
            system('apt-get', '-q', 'update')
                and Warsync::fault("something went wrong running apt-get update: $?");
        }
    }
    return 1;
}

1;