package Warsync::Server;

use strict;
use warnings;

use Warsync;

sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;

    unless (-r '/etc/warsync/server-key' and -r '/etc/warsync/client.list') {
        return undef;
    }

    my $config = Warsync::Config->new('/etc/warsync/client.list')
        or Warsync::fault("could not read client list: $!");
    
    return bless { config => $config } => $class;
}

sub apt_conf
{
    my $self = shift->new;

    return '' unless -e '/etc/apt/apt.conf';

    open(my $fh, '/etc/apt/apt.conf')
        or Warsync::fault("could not read /etc/apt/apt.conf");
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);

    return $slurp;
}

sub apt_preferences
{
    my $self = shift->new;

    return '' unless -e '/etc/apt/preferences';

    open(my $fh, '/etc/apt/preferences')
        or Warsync::fault("could not read /etc/apt/preferences");
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);

    return $slurp;
}

sub apt_sources_list
{
    my $self = shift->new;

    open(my $fh, '/etc/apt/sources.list')
        or Warsync::fault("could not read /etc/apt/sources.list");
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);

    return $slurp;
}

sub client
{
    my $self   = shift->new;
    my $client = shift;
    
    return undef unless $self->config->has_section("client-$client");
    
    return Warsync::Server::Client->new($self, $client);
}

sub config
{
    my $self = shift->new;
    
    return $self->{config};
}

sub list_clients
{
    my $self = shift->new;
    
    return map  { $_ =~ s/^client-//; $_ }
           grep { $_ =~ m/^client-/  } $self->{config}->list_sections;
}

sub name
{
    my $self = shift->new;

    return Sys::Hostname::hostname;
}

sub option
{
    my $self = shift->new;
    my $key  = shift;
    
    #print STDERR Data::Dumper::Dumper($self->{option});
    #print STDERR "option: $key\n";
    
    return $self->{options}{$key};
}

sub options
{
    my $self = shift->new;
    
    if (@_) {
        $self->{options} = shift;
        return $self;
    }
    
    return $self->{options};
}

sub packages
{
    my $self = shift->new;
    my $v    = shift;

    return Warsync::deep_copy(
        $self->{_packages} ||= Warsync::debian_packages($v)
    );
}

sub run
{
    my $self = shift->new;

    my $option = Warsync->options(
        'a|all-files',
        'client=s@',
        'debian',
        'debian-snapshot=s',
        'debian-from-snapshot=s',
    );

    $self->options($option);
    
    #print STDERR Data::Dumper::Dumper($self);
    #print STDERR 'dry-run: ', $self->option('n'), "\n";

    ## if --debian-snapshot argument given, do that.
    if ($self->option('debian-snapshot')) {
        Warsync::debian_snapshot($self->options);
        exit;
    }
    
    ## if --debian-from-snapshot argument given, do that.
    if ($self->option('debian-from-snapshot')) {
        Warsync::debian_from_snapshot($self->options);
        exit;
    }
    
    my $clients = $self->option('client') || [ $self->list_clients ];

    my @filepacks = @ARGV;

    if ($self->option('a')) {
        @filepacks = $self->filepacks;
    }

    ## they need to specify something to do.
    unless (@filepacks or $self->option('debian')) {
        Pod::Usage::pod2usage(1);
    }

    ## iterate through each client and replicate.
    foreach my $name (@$clients) {

        print "######################################################\n";

        my $client = $self->client($name);
        
        ## if we didn't get back a fqdn, this is a bad client name.
        unless ($client) {
            print STDERR $name, " is not a valid client\n";
            next;
        }

        my $dryrun = $self->option('n') ? ' (dry-run) ' : '';

        print "Replicating $name$dryrun...\n";

        if ($self->option('debian')) {
            $client->replicate_debian;
        }

        foreach my $filepack (@filepacks) {
            $client->replicate_filepack($filepack);
        }
    }
}

#############################################################################
#############################################################################

package Warsync::Server::Client;

## Warsync::Server::Client->new($server, $client_name);
sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;
    
    my $server = shift;
    my $name   = shift;
    
    $self = bless {
        server => $server,
        name   => $name,
        config => $server->config->get_section("client-$name")
    } => $class;
    
    return $self;
}

sub base_command
{
    my $self = shift->new;
    my $cmd  = shift;
    
    my $v = $self->server->option('v');
    
    
    my $rcmd = 'warsync-';
    $rcmd   .= 'compress-' if $self->compress;
    $rcmd   .= 'dry-run-'  if $self->server->option('n');
    $rcmd   .= 'quiet-'    if $self->server->option('q');
    $rcmd   .= "verbose$v-" if $v;
    
    $rcmd   .= $cmd;
    
    #print STDERR "cmd: $rcmd\n";
    
    return $rcmd;
}

sub command
{
    my $self   = shift->new;
    my $cmd    = $self->base_command(shift);
    my @args   = @_;

    my @scmd = ( 'ssh', '-p', $self->port,
                        ($self->compress ? '-C' : () ),
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-T', $self->fqdn, $cmd, @args );

    open(my $ssh_h, '-|') || exec( @scmd )
        or die "error '$!' trying to exec: @scmd";

    return $ssh_h;
}

sub gui_command
{
    my $self   = shift->new;
    
    ## run as non-gui command if this is not a tty.
    unless (-t) {
        my $comh = $self->command(@_);
        print <$comh>;
        return close($comh)
    }

    my $cmd    = $self->base_command(shift);
    my @args   = @_;
    
    my @scmd = ( 'ssh', '-p', $self->port,
                        ($self->compress ? '-C' : () ),
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-qt', $self->fqdn, $cmd, @args );

    system( @scmd ) and return undef;
    
    return 1;
}

sub compress
{
    my $self = shift->new;
    
    if (@_) {
        $self->set_key('compress', shift);
        return $self;
    }
    
    return $self->get_key('compress');
}    

sub config
{
    my $self = shift->new;
    
    return $self->{config};
}

sub ignore_debs
{
    my $self   = shift->new;

    my @debs = split ' ', $self->get_key('ignore_debs') || '';
    return \@debs;
}

sub fqdn
{
    my $self = shift->new;
    
    if (@_) {
        $self->set_key('fqdn', Warsync::is_fqdn(shift));
        return $self;
    }
    
    return $self->get_key('fqdn');
}

sub get_key
{
    my $self = shift->new;
    
    return $self->config->get_key(shift);
}

sub set_key
{
    my $self = shift->new;
    
    return $self->config->set_key(shift, shift);
}

sub name
{
    my $self = shift->new;
    
    return $self->{name};
}

sub packages
{
    my $self = shift->new;

    my $cmd_h = $self->command('echo-package-list');
    my $cpkgs = Warsync::parse_package_list($cmd_h);
    close($cmd_h);
    
    return $cpkgs;
}

sub port
{
    my $self = shift->new;
    
    if (@_) {
        my ($port) = shift =~ /(\d+)/;
        $self->set_key('port', $port);
        return $self;
    }
    
    return $self->get_key('port') || 22;
}

sub replicate_debian
{
    my $self   = shift->new;
    
    my $server = $self->server;
    
    print "Debian Package Syncronization\n";
    print "* determining server packages...";
    my $spkgs = $server->packages(1); print "done.\n";
    print "* determining client packages...";
    my $cpkgs = $self->packages(1); print "done.\n";
    my $ipkgs = $self->ignore_debs;
    
    print "* calculating differences...";
    my ($install, $remove, $purge) =
        Warsync::debian_package_diff($spkgs, $cpkgs, $ipkgs, 1);
    
    ## return if there is nothing to do.
    unless (@$install || @$remove || @$purge) {
        print "client up-to-date.\n";
        return;
    }

    ## have client update package sources.
    print "client out-of-sync.\n";
    print "* syncronizing package sources...\n";
    
    $self->gui_command("sync-package-sources")
            or Warsync::fault("client sync-package-sources returned status: $?");

    print "* installing/removing packages on client with apt-get...\n";

    my @apt = "package-install";
    push @apt, @$install; ## packages to install
    push @apt, map { "$_-" } @$remove;              ## packages to remove
    push @apt, map { "$_-" } @$purge;               ## packages to purge

    $self->gui_command(@apt)
            or Warsync::fault("client apt-get-install returned status: $?");
    
    if (@$purge) {
        print "* purging packages on client with apt-get...\n";
        my @apt = ("package-purge", @$purge);
        $self->gui_command(@apt)
                or Warsync::fault("client apt-get-purge returned status: $?");
    }

    print "* package syncronization completed!\n";
}

sub server
{
    my $self = shift->new;
    
    return $self->{server};
}

1;